#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <Utils/convertor.h>
#include <Degradations/BleedThrough.hpp>
#include <Degradations/BlurFilter.hpp>
#include <Degradations/GrayscaleCharsDegradationModel.hpp>
#include <Degradations/HoleDegradation.hpp>
#include <Degradations/PhantomCharacter.hpp>
#include <Degradations/ShadowBinding.hpp>
#include <chrono>
#include <ctime>


using namespace std;
using namespace cv;

string separator =
#ifdef _WIN32
    string("\\");
#else
    string("/");
#endif


// Code factorization of the exit failure process
int exitError (string degradation, int count){
    fprintf(stderr, "Error test %s n°%d (C++ function)\n", degradation.c_str(), count);
    return 1;
}

// Code factorization of the save image process
void saveImg (string degradation, int count, string outputDir, Mat outImg){
    string name = degradation + to_string(count) + ".png";
    imwrite(outputDir + separator + name, outImg);
    printf("%s (C++) has been generated without errors\n", name.c_str());
}



// ==================== Degradation tests ====================


int testGrayscaleCharsDegradationModel (string degradation, string referenceImageName, string outputDir){

    int count = 1;
    const int nbTests1 = 5; // Number of degradate_cv (with args) tests
    const int nbTests2 = 3; // Number of degradateByLevel_cv tests

    Mat img = imread(referenceImageName);
    GrayscaleCharsDegradationModel gdm = GrayscaleCharsDegradationModel(img);
    Mat outImg;

    /* Tests :
        degradate_cv() without arguments
    */

    try{
        outImg = gdm.degradate_cv();
        saveImg (degradation, count, outputDir, outImg);
        ++count;
    } catch (...) {
        return exitError(degradation, count);
    }


    /* Tests :
        degradate_cv with intensity level at 1
        degradate_cv with intensity level at an high value : 50
        degradate_cv with disconnection points at 100% (0% for the others)
        degradate_cv with overlapping points at 100% (0% for the others)
        degradate_cv with independant points at 100% (0% for the others)
    */

    int level [nbTests1] = {1, 50, 10, 10, 10};
    int independantPoints [nbTests1] = {33, 33, 0, 0, 100};
    int overlappingPoints [nbTests1] = {33, 33, 0, 100, 0};
    int disconnectionPoints [nbTests1] = {34, 34, 100, 0, 0};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

    for (int i = 0 ; i < nbTests1 ; ++i){
        try{
            gdm = GrayscaleCharsDegradationModel(img);
            outImg = gdm.degradate_cv(
                level[i],
                independantPoints[i],
                overlappingPoints[i],
                disconnectionPoints[i]
            );
            saveImg (degradation, count, outputDir, outImg);
            ++count;
        } catch (...){
            return exitError(degradation, count);
        }
    }


   /* Tests :
        degradateByLevel_cv with intensity level at a regular value : 10
        degradateByLevel_cv with intensity level at a minimum value : 1
        degradateByLevel_cv with intensity level at a high value : 50
   */

    int level2 [nbTests2] = {10, 1, 50};

    for (int i = 0 ; i < nbTests2 ; ++i){
        try{
            gdm = GrayscaleCharsDegradationModel(img);
            outImg = gdm.degradateByLevel_cv(level2[i]);
            saveImg (degradation, count, outputDir, outImg);
            ++count;
        } catch (...){
            return exitError(degradation, count);
        }
    }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

    return 0;

}


int testShadowBinding (string degradation, string referenceImageName, string outputDir){

    int count = 1;
    const int nbTests = 10;

    Mat img = imread(referenceImageName);
    QImage qImg;
    Mat outImg;

    /* Tests :
        shadowBinding() with the border type 0 (top)
        shadowBinding() with the border type 1 (right)
        shadowBinding() with the border type 2 (bottom)
        shadowBinding() with the border type 3 (left)
        shadowBinding() with intensity at the medium value : 0.5 and an angle of 30°
        shadowBinding() with intensity at the minimum value : 0 and an angle at the minimum value : 0°
        shadowBinding() with intensity at the maximum value : 1 and an angle at the maximum value : 90°
        shadowBinding() with a distance (size in pixel) of 0
        shadowBinding() with a distance (size in pixel) of 1
        shadowBinding() with a distance (size in pixel) of 1000
    */

    int border [nbTests] = {0, 1, 2, 3, 0, 0, 0, 0, 0, 0};
    int distance [nbTests] = {200, 200, 200, 200, 200, 200, 200, 0, 1, 1000};
    float intensity [nbTests] = {0.5, 0.5, 0.5, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5};
    float angle [nbTests] = {45, 45, 45, 45, 30, 0, 90, 45, 45, 45};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

    for (int i = 0 ; i < nbTests ; ++i){
        try{
            qImg = Convertor::getQImage(img);
            qImg = shadowBinding(qImg, (ShadowBorder)border[i], distance[i], intensity[i], angle[i]);
            outImg = Convertor::getCvMat(qImg);
            saveImg(degradation, count, outputDir, outImg);
            ++count;
        } catch (...) {
            return exitError(degradation, count);
        }
   }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

   return 0;

}


int testHoleDegradation (string degradation, string referenceImage1Name, string outputDir, string referenceImage2Name, string patternImg){

    int count = 1;
    const int nbTests1 = 4;
    const int nbTests2 = 2;

    Mat img1 = imread(referenceImage1Name);
    Mat img2 = imread(referenceImage2Name);
    Mat imgPattern = imread(patternImg);
    Mat outImg;

    int xOrigin = 50;
    int yOrigin = 50;
    int size = 1;

    /* Tests :
        holeDegradation() with 1st hole pattern and a colored(red) hole, situated at the top of the image
        holeDegradation() with 2nd hole pattern and a grey hole, situated at the right of the image
        holeDegradation() with 3rd hole pattern and a black hole, situated at the bottom of the image
        holeDegradation() with 4th hole pattern and a white hole, situated at the left of the image
    */

    int holetype [nbTests1] = {0, 1, 2, 3};
    int side [nbTests1] = {0, 1, 2, 3};
    QColor color [nbTests1] = {{255, 0, 0}, {128, 128, 128}, {0, 0, 0}, {255, 255, 255}};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

    for (int i = 0 ; i < nbTests1 ; ++i){
        try{
            outImg = holeDegradation(img1, imgPattern, xOrigin, yOrigin, size, (HoleType)holetype[i], side[i], color[i]);
            saveImg(degradation, count, outputDir, outImg);
            ++count;
        } catch (...) {
            return exitError(degradation, count);
        }
    }


    /* Tests :
        holeDegradation() with an image in the hole, width at 50, and intensity at minimum value : 0
        holeDegradation() with an image in the hole, width at 100, and intensity at half the default value : 500
    */

    int holetype2 [nbTests2] = {1, 1};
    int side2 [nbTests2] = {1, 1};
    QColor color2 [nbTests2] = {{128, 128, 128}, {128, 128, 128}};
    int width2 [nbTests2] = {50, 100};
    float intensity2 [nbTests2] = {0, 500};

    for (int i = 0 ; i < nbTests2 ; ++i){
        try{
            outImg = holeDegradation(img1, imgPattern, xOrigin, yOrigin, size, (HoleType)holetype2[i], side2[i], color2[i], img2, width2[i], intensity2[i]);
            saveImg(degradation, count, outputDir, outImg);
            ++count;
        } catch (...) {
            return exitError(degradation, count);
        }
    }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

    return 0;

}


int testBleedThrough (string degradation, string referenceImage1Name, string outputDir, string referenceImage2Name){

    int count = 1;
    const int nbTests = 4;

    Mat img1 = imread(referenceImage1Name);
    Mat img2 = imread(referenceImage2Name);
    QImage qImg1;
    QImage qImg2 = Convertor::getQImage(img2);
    Mat outImg;

    int nbThreads = -1;

    /* Tests :
        bleedThrough() with 50 iterations
        bleedThrough() with the minimum amount of iterations : 0
        bleedThrough() with a high amount of iterations : 100
        bleedThrough() with the verso image positionned at (x=100, y=100) in the recto image frame
    */

   int nbIter [nbTests] = {50, 0, 100, 50};
   int pos [nbTests][2] = {{0, 0}, {0, 0}, {0, 0}, {100, 100}};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

   for (int i = 0 ; i < nbTests ; ++i){
       try{
           qImg1 = Convertor::getQImage(img1);
           qImg1 = bleedThrough(qImg1, qImg2, nbIter[i], pos[i][0], pos[i][1], nbThreads);
           outImg = Convertor::getCvMat(qImg1);
           saveImg(degradation, count, outputDir, outImg);
           ++count;
       } catch (...) {
           return exitError(degradation, count);
       }
   }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

   return 0;

}


int testPhantomCharacter (string degradation, string referenceImageName, string outputDir, char * patternDir){

    int count = 1;
    const int nbTests = 3;

    Mat img = imread(referenceImageName);
    Mat outImg;

    /* Tests :
        phantomCharacter() with frequency at 15% (rare)
        phantomCharacter() with frequency at 40% (frequent)
        phantomCharacter() with frequency at 70% (very frequent)
    */

   int frequency [nbTests] = {0, 1, 2};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

   for (int i = 0 ; i < nbTests ; ++i){
       try{
           outImg = phantomCharacter(img, (Frequency)frequency[i], patternDir);
           saveImg(degradation, count, outputDir, outImg);
           ++count;
       } catch (...) {
           return exitError(degradation, count);
       }
   }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

   return 0;

}


int testBlurFilter (string degradation, string referenceImageName, string outputDir){

    int count = 1;
    const int nbTests = 9;

    Mat img = imread(referenceImageName);
    Mat outImg;

    /* Tests :
        blurFilter() with method : Gaussian and medium intensity : 11
        blurFilter() with method : Gaussian and minimum intensity : 1
        blurFilter() with method : Gaussian and maximum intensity : 21
        blurFilter() with method : Median and medium intensity : 11
        blurFilter() with method : Median and minimum intensity : 1
        blurFilter() with method : Median and maximum intensity : 21
        blurFilter() with method : Normal and medium intensity : 11
        blurFilter() with method : Normal and minimum intensity : 1
        blurFilter() with method : Normal and maximum intensity : 21
    */

    int method [nbTests] = {0, 0, 0, 1, 1, 1, 2, 2, 2};
    int intensity [nbTests] = {11, 1, 21, 11, 1, 21, 11, 1, 21};

    chrono::time_point<chrono::system_clock> start, end;
    start = chrono::system_clock::now();

    for (int i = 0 ; i < nbTests ; ++i){
        try{
            outImg = blurFilter(img, (Method)method[i], intensity[i]);
            saveImg(degradation, count, outputDir, outImg);
            ++count;
        } catch (...) {
            return exitError(degradation, count);
        }
    }

    end = chrono::system_clock::now();
    chrono::duration<double> elapsed_seconds = end - start;
    cout << "Time C++ : " << elapsed_seconds.count() << "s\n";

    return 0;

}



int main(int argc, char ** argv){

    string degradation = argv[1];
    string referenceImageName = argv[2];
    string outputDir = argv[3];


    if (degradation == "shadowbinding"){
        return testShadowBinding(degradation, referenceImageName, outputDir);
    }
    else if(degradation == "grayscale"){
        return testGrayscaleCharsDegradationModel(degradation, referenceImageName, outputDir);
    }
    else if (degradation == "holedegradation"){
        string referenceImage2Name = argv[4];
        string patternImg = argv[5];
        return testHoleDegradation(degradation, referenceImageName, outputDir, referenceImage2Name, patternImg);
    }
    else if (degradation == "bleedthrough"){
        string referenceImage2Name = argv[4];
        return testBleedThrough(degradation, referenceImageName, outputDir, referenceImage2Name);
    }
    else if (degradation == "phantomcharacter"){
        char * patternDir = argv[4];
        return testPhantomCharacter(degradation, referenceImageName, outputDir, patternDir);
    }
    else if (degradation =="blurfilter"){
        return testBlurFilter(degradation, referenceImageName, outputDir);
    }
    else{
        fprintf(stderr, "Unknown degradation\n");
        fprintf(stderr, "This program is not intended to be used as a standalone, please use the Python test first\n");
    }

}
