# -*- coding: utf-8 -*-

import cv2
import sys
import os
import time

sys.path.insert(0, '../wrapper/python')
import DocCreator

cprog = "./testDegradationWrapper "


# Compare pixel by pixel two given images
def compareImage (img1Name, img2Name):

    img1 = cv2.imread(img1Name)
    img2 = cv2.imread(img2Name)

    if img1.shape == img2.shape : # Check dimensions
        diff = cv2.subtract(img1, img2)
        blue, green, red = cv2.split(diff)
        if (cv2.countNonZero(blue) == 0) and (cv2.countNonZero(green) == 0) and (cv2.countNonZero(red) == 0):
            return 0 # Same images

    return 1 # Different images


# Code factorization of the save image process
def saveImg (degradation, count, outputDir, outImg):
    name = degradation + str(count) + ".png"
    cv2.imwrite(os.path.join(outputDir, name), outImg)
    print(name + " (Python) has been generated without errors")


# Code factorization of the exit failure process
def exitError (degradation, count):
    print("Error testing " + degradation + " n°" + str(count) + " (Python function)")
    sys.exit(1)


# Code factorization of the comparison of all images in a folder
def compareDir (degradation, outputDir1, outputDir2):

    dir1List = os.listdir(outputDir1)
    dir2List = os.listdir(outputDir2)

    # Check that the given folders are not empty
    if not(len(dir1List) > 0 and len(dir2List) > 0):
        print("There must be at least one image in each folder to compare")
        print("There is " + str(len(dir1List)) + " images in " + outputDir1)
        print("There is " + str(len(dir2List)) + " images in " + outputDir2)
        sys.exit(1)

    # Check that both of the given folders contain the same number of images
    if ((len(dir1List) != len(dir2List))):
        print("Error : the two output folders are not containing the same number of images")
        print(outputDir1 + " has " + str(len(dir1List)) + " images")
        print(outputDir2 + " has " + str(len(dir2List)) + " images")
        sys.exit(1)

    size = len(dir1List)
    count = 0
    # The images are numbered by degradation, so we can compare them by alphabetical order
    dir1Sorted = sorted(dir1List)
    dir2Sorted = sorted(dir2List)

    # Call the compareImage function for each pair of images in the two folders
    for i in range (0, size):
        imgPython = os.path.join(outputDir1, dir1Sorted[i])
        imgC = os.path.join(outputDir2, dir2Sorted[i])
        result = compareImage(imgPython, imgC)
        if (result != 0):
            print("Error : " + imgPython + " and " + imgC + " are not identical")
        else:
            print(imgPython + " and " + imgC + " are identical")
            count += 1

    if (count == 0): # All pair of images are different
        print("Error : no images generated by " + degradation + " were identical")
        sys.exit(1)
    elif (count == size): # All pair of images are identical
        print("All " + degradation + " generated images are identical")
        sys.exit(0)
    else: # Print the number of different pair of images found
        print("Error : only " + str(count) + "/" + str(size) + " images generated by " + degradation + " are identical")
        sys.exit(1)


# Code factorization of the C++ functions calling process
def callCFunc (argList):
    cmd = cprog + argList
    try:
        # If the C++ functions didn't work for some reason, the return code will be != 0
        if (os.system(cmd) != 0):
            print("Error while generating C++ functions")
            sys.exit(1)
    except:
        print("Error while calling C++ functions")
        sys.exit(1)


# Code factorization for the creation/cleaning of the output folders
def manageDir (outputDir):
    if not(os.path.exists(outputDir)): # Create folder if it doesn't exists
        os.makedirs(outputDir)
    else: # Clean the given folder if it exists
        for f in os.listdir(outputDir):
            fPath = os.path.join(outputDir, f)
            try:
                if os.path.isfile(fPath):
                    os.unlink(fPath)
                elif os.path.isdir(fPath):
                    shutil.rmtree(fPath)
            except Exception as e:
                print("Error clearing folder : " + e)




# ==================== Degradation tests ====================


def testGrayscaleCharsDegradationModel (degradation, referenceImageName, outputDir1, outputDir2):

    count = 1
    img = cv2.imread(referenceImageName)

    """ Tests :
    grayscaleChars() without arguments
    """

    try:
        outImg = DocCreator.grayscaleChars(img)
        saveImg(degradation, count, outputDir1, outImg)
        count += 1
    except:
        exitError(degradation, count)


    """ Tests :
    grayscaleChars() with intensity at 1
    grayscaleChars() with intensity at a high level : 50
    grayscaleChars() with disconnection points at 100% (0% for the others)
    grayscaleChars() with overlapping points at 100% (0% for the others)
    grayscaleChars() with independant points at 100% (0% for the others)
    """

    level = [1, 50, 10, 10, 10]
    IPoints = [33, 33, 0, 0, 100]
    OPoints = [33, 33, 0, 100, 0]
    DPoints = [34, 34, 100, 0, 0]

    start = time.time()

    for i in range (0, len(level)):
        try:
            outImg = DocCreator.grayscaleChars(img, level[i], IPoints[i], OPoints[i], DPoints[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)


    """ Tests :
    grayscaleCharsByLevel() with intensity at a regular level : 10
    grayscaleCharsByLevel() with intensity at the minimum level : 1
    grayscaleCharsByLevel() with intensity at a high level : 50
    """

    level = [10, 1, 50]

    for i in range (0, len(level)):
        try:
            outImg = DocCreator.grayscaleCharsByLevel(img, level[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)

    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    argList = degradation + ' ' + referenceImageName + ' ' + outputDir2
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)


def testShadowBinding (degradation, referenceImageName, outputDir1, outputDir2):

    count = 1
    img = cv2.imread(referenceImageName)

    """ Tests :
    shadowBinding() with the border type 0 (top)
    shadowBinding() with the border type 1 (right)
    shadowBinding() with the border type 2 (bottom)
    shadowBinding() with the border type 3 (left)
    shadowBinding() with intensity at the medium value : 0.5 and an angle of 30°
    shadowBinding() with intensity at the minimum value : 0 and an angle at the minimum value : 0°
    shadowBinding() with intensity at the maximum value : 1 and an angle at the maximum value : 90°
    shadowBinding() with a distance (size in pixel) of 0
    shadowBinding() with a distance (size in pixel) of 1
    shadowBinding() with a distance (size in pixel) of 1000
    """

    border = [0, 1, 2, 3, 0, 0, 0, 0, 0, 0]
    distance = [200, 200, 200, 200, 200, 200, 200, 0, 1, 1000]
    intensity = [0.5, 0.5, 0.5, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5]
    angle = [45, 45, 45, 45, 30, 0, 90, 45, 45, 45]

    start = time.time()

    for i in range (0, len(border)):
        try:
            outImg = DocCreator.shadowBinding(img, border[i], distance[i], intensity[i], angle[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)


    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    argList = degradation + ' ' + referenceImageName + ' ' + outputDir2
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)


def testHoleDegradation (degradation, referenceImage1Name, outputDir1, outputDir2, referenceImage2Name, patternImg):

    count = 1
    img1 = cv2.imread(referenceImage1Name)
    img2 = cv2.imread(referenceImage2Name)
    pattern = cv2.imread(patternImg)

    xOrigin = 50
    yOrigin = 50
    size = 1

    """ Tests :
    holeDegradation() with 1st hole pattern and a colored(red) hole, situated at the top of the image
    holeDegradation() with 2nd hole pattern and a grey hole, situated at the right of the image
    holeDegradation() with 3rd hole pattern and a black hole, situated at the bottom of the image
    holeDegradation() with 4th hole pattern and a white hole, situated at the left of the image
    """

    holetype = [0, 1, 2, 3]
    side = [0, 1, 2, 3]
    color = [[255, 0, 0], [128, 128, 128], [0, 0, 0], [255, 255, 255]]

    start = time.time()

    for i in range (0, len(holetype)):
        try:
            outImg = DocCreator.holeDegradation(img1, pattern, xOrigin, yOrigin, size, holetype[i], side[i], color[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)


    """ Tests :
    holeDegradation() with an image in the hole, width at 50, and intensity at minimum value : 0
    holeDegradation() with an image in the hole, width at 100, and intensity at half the default value : 500
    """

    holetype = [1, 1]
    side = [1, 1]
    color = [[128, 128, 128], [128, 128, 128]]
    width = [50, 100]
    intensity = [0, 500]

    for i in range (0, len(holetype)):
        try:
            outImg = DocCreator.holeDegradation(img1, pattern, xOrigin, yOrigin, size, holetype[i], side[i], color[i], img2, width[i], intensity[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)

    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    mainArgList = degradation + ' ' + referenceImage1Name + ' ' + outputDir2
    additionalArgList = ' ' + referenceImage2Name + ' ' + patternImg
    argList = mainArgList + additionalArgList
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)


def testBleedThrough (degradation, referenceImage1Name, outputDir1, outputDir2, referenceImage2Name):

    count = 1
    img1 = cv2.imread(referenceImage1Name)
    img2 = cv2.imread(referenceImage2Name)

    nbThreads = -1

    """ Tests :
    bleedThrough() with 50 iterations
    bleedThrough() with the minimum amount of iterations : 0
    bleedThrough() with a high amount of iterations : 100
    bleedThrough() with the verso image positionned at (x=100, y=100) in the recto image frame
    """

    nbIter = [50, 0, 100, 50]
    pos = [[0, 0], [0, 0], [0, 0], [100, 100]]

    start = time.time()

    for i in range (0, len(nbIter)):
        try:
            outImg = DocCreator.bleedThrough(img1, img2, nbIter[i], pos[i][0], pos[i][1], nbThreads)
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)

    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    mainArgList = degradation + ' ' + referenceImage1Name + ' ' + outputDir2
    additionalArgList = ' ' + referenceImage2Name
    argList = mainArgList + additionalArgList
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)


def testPhantomCharacter (degradation, referenceImageName, outputDir1, outputDir2, patternDir):

    count = 1
    img = cv2.imread(referenceImageName)

    """ Tests :
    phantomCharacter() with frequency at 15% (rare)
    phantomCharacter() with frequency at 40% (frequent)
    phantomCharacter() with frequency at 70% (very frequent)
    """

    frequency = [0, 1, 2]

    start = time.time()

    for i in range (0, len(frequency)):
        try:
            outImg = DocCreator.phantomCharacter(img, frequency[i], patternDir)
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)

    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    mainArgList = degradation + ' ' + referenceImageName + ' ' + outputDir2
    additionalArgList = ' ' + patternDir
    argList = mainArgList + additionalArgList
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)


def testBlurFilter (degradation, referenceImageName, outputDir1, outputDir2):

    count = 1
    img = cv2.imread(referenceImageName)

    """ Tests :
    blurFilter() with method : Gaussian and medium intensity : 11
    blurFilter() with method : Gaussian and minimum intensity : 1
    blurFilter() with method : Gaussian and maximum intensity : 21
    blurFilter() with method : Median and medium intensity : 11
    blurFilter() with method : Median and minimum intensity : 1
    blurFilter() with method : Median and maximum intensity : 21
    blurFilter() with method : Normal and medium intensity : 11
    blurFilter() with method : Normal and minimum intensity : 1
    blurFilter() with method : Normal and maximum intensity : 21
    """

    method = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    intensity = [11, 1, 21, 11, 1, 21, 11, 1, 21]

    start = time.time()

    for i in range (0, len(method)):
        try:
            outImg = DocCreator.blurFilter(img, method[i], intensity[i])
            saveImg(degradation, count, outputDir1, outImg)
            count += 1
        except:
            exitError(degradation, count)

    end = time.time()
    print("Time Python : %.6f%s" % ((end - start), "s"))

    # Calling C++ function
    argList = degradation + ' ' + referenceImageName + ' ' + outputDir2
    callCFunc(argList)

    compareDir(degradation, outputDir1, outputDir2)



def main (degradation, referenceImage1Name, outputDir1, outputDir2, referenceImage2Name, patternImg, patternDir):

    outputDir1 = os.path.join(outputDir1, degradation)
    outputDir2 = os.path.join(outputDir2, degradation)

    manageDir(outputDir1)
    manageDir(outputDir2)

    if (degradation == 'shadowbinding'):
        testShadowBinding(degradation, referenceImage1Name, outputDir1, outputDir2)
    elif (degradation == 'grayscale'): # Contains random, must compile with -DWITHOUT_RANDOM
        testGrayscaleCharsDegradationModel(degradation, referenceImage1Name, outputDir1, outputDir2)
    elif (degradation == 'holedegradation'):
        testHoleDegradation(degradation, referenceImage1Name, outputDir1, outputDir2, referenceImage2Name, patternImg)
    elif (degradation == 'bleedthrough'):
        testBleedThrough(degradation, referenceImage1Name, outputDir1, outputDir2, referenceImage2Name)
    elif (degradation == 'phantomcharacter'): # Contains random, must compile with -DWITHOUT_RANDOM
        testPhantomCharacter(degradation, referenceImage1Name, outputDir1, outputDir2, patternDir)
    elif (degradation == 'blurfilter'):
        testBlurFilter(degradation, referenceImage1Name, outputDir1, outputDir2)
    else:
        print("Unknown degradation, please use one of the following : ")
        print("grayscale (test will not pass until the random seed is set), shadowbinding, holedegradation, bleedthrough, phantomcharacter (idem grayscale), blurfilter")
        sys.exit(1)




if __name__ == "__main__":
    if (len(sys.argv) < 8):
        print("Error launching tests, not enough arguments. Argument list must contain :")
        print("degradation")
        print("referenceImage1")
        print("outputDir1")
        print("outputDir2")
        print("referenceImage2")
        print("patternImg")
        print("patternDir")
        sys.exit(1)
    main(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6], sys.argv[7])
