#include "Degradations.hpp"
#include <Utils/convertor.h> //B: pour QImage (pour BleedThrough)

#include <opencv2/imgproc/imgproc.hpp> //cvtColor

/************************************
 *  GrayscaleCharsDegradationModel  *
 ************************************/
void grayscaleChars(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                    uint8_t *imgOut, int imgOutLength,
                    int level, float I, float O, float D)

{
    // Check that the image isn't empty
    // The degradations uses input images in 3 channels (UC3)
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);

    // The output image should have the same number of values as the input image
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    // We start by passing the array to a cv::Mat so we can use it later
    cv::Mat img(nbRows, nbCols, CV_8UC3, imgIn);

    // We call the degradation
    GrayscaleCharsDegradationModel gdm(img);
    cv::Mat res = gdm.degradate_cv(level, I, O, D);

    // If needed we convert the image back to the input format before returning it
    cvtColor(res, res, cv::COLOR_GRAY2RGB);
    // Write the result to imgOut, this is the returned image
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}

void grayscaleCharsByLevel(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                           uint8_t *imgOut, int imgOutLength,
                           int level)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    cv::Mat img(nbRows, nbCols, CV_8UC3, imgIn);

    GrayscaleCharsDegradationModel gdm(img);
    cv::Mat res = gdm.degradateByLevel_cv(level);

    cvtColor(res, res, cv::COLOR_GRAY2RGB);
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}


/*******************
 *  ShadowBinding  *
 *******************/
void shadowBinding (uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                    uint8_t *imgOut, int imgOutLength,
                    int border, int distance, float intensity, float angle)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    cv::Mat img(nbRows, nbCols, CV_8UC3, imgIn);

    // We don't want the original data to be modified so we degrade a clone of it
    cv::Mat res = img.clone();
    shadowBinding(res, (ShadowBorder) border, distance, intensity, angle);

    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}


/*********************
 *  HoleDegradation  *
 *********************/
void holeDegradation(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                     uint8_t *imgOut, int imgOutLength,
                     uint8_t *pattern, int nbRowsPattern, int nbColsPattern, int nbChannelsPattern,
                     int xOrigin, int yOrigin, int size, int holetype, int side, int color[3],
                     uint8_t *imgBelow, int nbRowsBelow, int nbColsBelow, int nbChannelsBelow,
                     int width, float intensity)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    cv::Mat im_top(nbRows, nbCols, CV_8UC3, imgIn);
    cv::Mat im_below(nbRowsBelow, nbColsBelow, CV_8UC3, imgBelow);
    cv::Mat im_pattern(nbRowsPattern, nbColsPattern, CV_8UC3, pattern);

    cv::Mat res = holeDegradation(im_top,
                                  im_pattern,
                                  xOrigin,
                                  yOrigin,
                                  size,
                                  (HoleType) holetype,
                                  side,
                                  QColor(color[0], color[1], color[2], 1),
                                  im_below,
                                  width,
                                  intensity);

    cvtColor(res, res, cv::COLOR_RGBA2RGB);
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}


/******************
 *  BleedThrough  *
 ******************/
void bleedThrough(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                  uint8_t *imgOut, int imgOutLength,
                  uint8_t *imgBelow, int nbRowsBelow, int nbColsBelow, int nbChannelsBelow,
                  int nbIter, int x, int y, int nbThreads)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    cv::Mat imgRecto(nbRows, nbCols, CV_8UC3, imgIn);
    cv::Mat imgVerso(nbRowsBelow, nbColsBelow, CV_8UC3, imgBelow);
    QImage qImgRecto = Convertor::getQImage(imgRecto);
    QImage qImgVerso = Convertor::getQImage(imgVerso);

    QImage qRes = bleedThrough(qImgRecto, qImgVerso, nbIter, x, y, nbThreads);

    cv::Mat res = Convertor::getCvMat(qRes);
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}


/**********************
 *  PhantomCharacter  *
 **********************/
void phantomCharacter(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                      uint8_t *imgOut, int imgOutLength,
                      int frequency, char *phantomPatternsPath)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    cv::Mat img(nbRows, nbCols, CV_8UC3, imgIn);

    cv::Mat res = phantomCharacter(img, (Frequency) frequency, (QString) phantomPatternsPath);
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}


/****************
 *  BlurFilter  *
 ****************/
void blurFilter(uint8_t *imgIn, int nbRows, int nbCols, int nbChannels,
                uint8_t *imgOut, int imgOutLength,
                int method, int intensity)
{
    assert(imgIn && nbRows > 0 && nbCols > 0 && nbChannels == 3);
    assert(imgOutLength == nbRows * nbCols * nbChannels);

    if (method < 2 && intensity%2 == 0) {
        fprintf(stderr, "Method %d requires odd intensity value (current: %d)\n", method, intensity);
        exit(1);
    }
    cv::Mat img(nbRows, nbCols, CV_8UC3, imgIn);

    cv::Mat res = blurFilter(img, (Method) method, intensity);
    uint8_t* dataPointer = reinterpret_cast<uint8_t*>(res.data);
    memcpy(imgOut, dataPointer, imgOutLength * sizeof(uint8_t));
}
