#include "Generation.hpp"
#include "Generation_aux.cpp"
#include <iomanager/fontfilemanager.h>
#include <QFileInfo>

using namespace std;

Generation::Generation() {
  initConfigManager();
}

QImage loadBackground(QString bgName) {
    /* Check given path for background */
    QFileInfo check_file(bgName);
    if(check_file.exists() && check_file.isFile()) {
        return QImage(bgName);
    }

    /* Check if the background is in /data */
    QString dataBgPath = Core::ConfigurationManager::get(AppConfigMainGroup,
                                                         AppConfigBackgdFolderKey)
                                                         .toString();
    Context::BackgroundContext *bgContext = Context::BackgroundContext::instance();
    bgContext->initialize(dataBgPath);

    QString bgPath = Context::BackgroundContext::instance()->getPath() + bgName;
    QFileInfo check_data_file(bgPath);
    if( !(check_data_file.exists() && check_data_file.isFile()) )  {
        qDebug() << "The background" << bgName << "does not exist\n";
        abort();
    }

    return QImage(bgPath);
}

Models::Font * loadFont(QString fontName) {
    /* Check given path for font */
    QFileInfo check_file(fontName);
    if(check_file.exists() && check_file.isFile()) {
        return IOManager::FontFileManager::fontFromXml(fontName);
    }

    /* Load font from /data */
    QString dataFontPath = Core::ConfigurationManager::get(AppConfigMainGroup,
                                                           AppConfigFontFolderKey)
                                                           .toString();
    Context::FontContext *fontContext = Context::FontContext::instance();
    fontContext->initialize(dataFontPath, "of");

    return fontContext->getFont(fontName);
}

void Generation::create(int createFunctionID,
                        const string &outputFolderPath,
                        const vector<string> &backgroundList,
                        const vector<string> &fontList,

                        const string &txtDirPath,

                        int bottomMarginMin,
                        int bottomMarginMax,
                        int leftMarginMin,
                        int leftMarginMax,
                        int rightMarginMin,
                        int rightMarginMax,
                        int topMarginMin,
                        int topMarginMax)
{
    if(createFunctionID < 0 || createFunctionID > 3){
        std::cerr << "createFunctionID must be between 0 and 3\n";
        return ;
    }

    int pageMarginTop = RandomElement().randomInt(topMarginMin, topMarginMax);
    int pageMarginBottom = RandomElement().randomInt(bottomMarginMin, bottomMarginMax);
    int pageMarginLeft = RandomElement().randomInt(leftMarginMin, leftMarginMax);
    int pageMarginRight = RandomElement().randomInt(rightMarginMin, rightMarginMax);

    DocumentGeneration docGen;

    docGen._txtDir = QString::fromStdString(txtDirPath);
    docGen._outputFolder = QString::fromStdString(outputFolderPath);
    docGen._pageMarginTop = pageMarginTop;
    docGen._pageMarginLeft = pageMarginLeft;
    docGen._pageMarginRight = pageMarginRight;
    docGen._pageMarginBottom = pageMarginBottom;

    using param_type = typename std::uniform_int_distribution<int>::param_type;
    static std::mt19937 eng {std::random_device{}()};
    static std::uniform_int_distribution<int> dist;

    //create all documents with one random font and one random background.
    if(createFunctionID == 0){
        int fontId = dist(eng, param_type{0, (int) fontList.size() - 1});
        QString fontName = QString::fromStdString(fontList[fontId]);
        docGen._font = loadFont(fontName);

        int bgId = dist(eng, param_type{0, (int) backgroundList.size() - 1});
        QString bgName = QString::fromStdString(backgroundList[bgId]);
        docGen._bgName = bgName;
        docGen._bg = loadBackground(bgName);

        docGen._lineSpacing = computeBestLineSpacing(*docGen._font);

        docGen.generate();
    }

    //create all documents with one random background and all fonts.
    else if(createFunctionID == 1){
        int bgId = dist(eng, param_type{0, (int) backgroundList.size() - 1});
        QString bgName = QString::fromStdString(backgroundList[bgId]);
        docGen._bgName = bgName;
        docGen._bg = loadBackground(bgName);

        for(int fontId = 0; fontId < (int) fontList.size(); fontId++){
            docGen._font = loadFont(QString::fromStdString(fontList[fontId]));
            docGen._lineSpacing = computeBestLineSpacing(*docGen._font);
            docGen.generate();
        }
    }

    //create all documents with one random font and all backgrounds.
    else if(createFunctionID == 2){
        int fontId = dist(eng, param_type{0, (int) fontList.size() - 1});
        docGen._font = loadFont(QString::fromStdString(fontList[fontId]));
        docGen._lineSpacing = computeBestLineSpacing(*docGen._font);

        for(int bgId = 0; bgId < (int) backgroundList.size(); bgId++){
            QString bgName = QString::fromStdString(backgroundList[bgId]);
            docGen._bgName = bgName;
            docGen._bg = loadBackground(bgName);
            docGen.generate();
        }
    }

    //create all documents with all combinations of backgrounds and fonts.
    else if(createFunctionID == 3){
        for(int fontId = 0; fontId < (int) fontList.size(); fontId++){
            for(int bgId = 0; bgId < (int) backgroundList.size(); bgId++){

                QString bgName = QString::fromStdString(backgroundList[bgId]);

                docGen._font = loadFont(QString::fromStdString(fontList[fontId]));
                docGen._bgName = bgName;
                docGen._bg = loadBackground(bgName);

                docGen._lineSpacing = computeBestLineSpacing(*docGen._font);

                docGen.generate();
            }
        }
    }

}