import ply.yacc as yacc
import ply.lex as lex
import listeDegradation as ld


class lexYaccAnalyse(object):

    # Builder
    def __init__(self):
        self.tabDegradation = None
        self.list = ld.listeDegradation()
        self.error = 0
        self.firstStringError =""
        self.lastStringError=""
        self.numLignError = 0


    # === Lex part ===

    # Tokens initialization (words from the language)
    tokens = (
        'AND',
        'IMPLICATION',
        'OPENPAREN',
        'CLOSEPAREN',
        'REPERTORY',
        'FONCTION',
        'PARAM',
        'GENERATION',
    )

    # Tokens definition

    # Spaces and linebreaks are ignored
    t_ignore = '[\t\n]+'

    # def t_NOM_TOKEN(self, t):
    def t_AND(self, t):
        # Token regular expression
        r'&&'
        # Chain counter
        t.lexer.lineno += len(t.value)
        # Return for the yacc value
        return t

    def t_IMPLICATION(self, t):
        r'=>'
        t.lexer.lineno += len(t.value)
        return t

    def t_OPENPAREN(self, t):
        r'\('
        t.lexer.lineno += len(t.value)
        return t

    def t_CLOSEPAREN(self, t):
        r'\)'
        t.lexer.lineno += len(t.value)
        return t

    def t_FONCTION(self, t):
        r'grayscaleChars|shadowBinding|blurFilter|holeDegradation|bleedThrough|phantomCharacter'
        t.lexer.lineno += len(t.value)
        return t

    def t_GENERATION(self, t):
        r'create'
        t.lexer.lineno += len(t.value)
        return t

    def t_REPERTORY(self, t):
        r'r:/{0,1}(([a-zA-Z0-9]|\.|/|:|-|_)*/{0,1})+'
        t.lexer.lineno += len(t.value)
        return t

    def t_PARAM(self, t):
        r'([a-zA-Z0-9]|\.|/|:|-|_)+(,([a-zA-Z0-9]|\.|/|:|-|_)+)*'
        t.lexer.lineno += len(t.value)
        return t

    # Unknown parameter (ex.: @)
    def t_error(self, t):
        self.error = 1
        if (self.firstStringError == ""):
            self.firstStringError = t.value
        self.lastStringError = t.value
        t.lexer.skip(1)

    # Lexer construction
    def buildLexer(self):
        self.lexer = lex.lex(module=self)

    # Lexer test
    def testLexer(self, data):
        self.lexer.input(data)
        while True:
            tok = self.lexer.token()
            if not tok:
                break
            (tok)


    # Yacc parser

    # Initial state
    def p_start(self, p):
        '''start    : repertoryIn IMPLICATION expr
                    | generation IMPLICATION expr
                    | generation IMPLICATION end
        '''


    # State : Repertory : writes the directory name in the structure
    def p_repertoryIn(self, p):
        '''repertoryIn : REPERTORY
        '''
        self.list.newLign()
        self.list.addRepertory(p[1])

    def p_repertoryOut(self, p):
        '''repertoryOut : REPERTORY
        '''
        self.list.addRepertory(p[1])


    # State : Generation : writes generation function name with params in the structure
    def p_generation(self, p):
        '''generation : GENERATION OPENPAREN PARAM CLOSEPAREN
        '''
        self.list.newLign()
        self.list.addGeneration(p[1], p[3])

    # State : function : writes the degradation function name in the structure
    def p_fonction(self, p):
        '''fonction   : FONCTION OPENPAREN PARAM CLOSEPAREN
        '''

        self.list.addFonction(p[1], p[3])

    def p_expr(self, p):
        '''expr     : fonction IMPLICATION expr
                    | fonction IMPLICATION end
        '''

    def p_end(self, p):
        '''end      : repertoryOut
                    | repertoryOut AND start
        '''


    def p_error(self, p):
        self.error = 1
        self.numLignError = self.lexer.lineno - 1

    def buildParser(self):
        self.parser = yacc.yacc(module=self)

    def testParser(self, data):
        self.parser.parse(data)
        if (self.error == 1):
            # Getter for an unrecognized state
            lenghtError = len(self.firstStringError) - len(self.lastStringError)
            lenghtError += 1
            stringError = self.firstStringError[0:lenghtError]
            # Isolation of the first error if multiple
            stringError = stringError.split('=>')[0]
            stringError = stringError.split('&&')[0]
            stringError = stringError.split('(')[0]
            stringError = stringError.split(')')[0]
            messageError = "char num " + str(self.numLignError) + ", state unknown : " + stringError
            return messageError
        return self.error

    def getList(self):
        return self.list


# TEST
# m = lexYaccAnalyse()
# m.buildLexer()           # Build the lexer
# m.buildParser()

# print("Resultat")
# res = m.testParser("generation1(999) => grayScale(param1) => /rep2")
# print(res)
